@@ -1,6 +1,6 @@
 from __future__ import absolute_import, division, print_function
 
-from collections import deque
+from collections import deque, defaultdict
 from datetime import timedelta
 import functools
 import logging
@@ -8,6 +8,7 @@ import six
 import sys
 import threading
 from time import time
+from typing import Any, Callable, Hashable, Union
 import weakref
 
 import toolz
@@ -20,15 +21,18 @@ try:
 except ImportError:
     PollIOLoop = None  # dropped in tornado 6.0
 
+try:
+    from distributed.client import default_client as _dask_default_client
+except ImportError:  # pragma: no cover
+    _dask_default_client = None
+
 from collections.abc import Iterable
 
-from .compatibility import get_thread_identity
+from threading import get_ident as get_thread_identity
 from .orderedweakset import OrderedWeakrefSet
 
 no_default = '--no-default--'
 
-_global_sinks = set()
-
 _html_update_streams = set()
 
 thread_state = threading.local()
@@ -43,6 +47,15 @@ def get_io_loop(asynchronous=None):
     if asynchronous:
         return IOLoop.current()
 
+    if _dask_default_client is not None:
+        try:
+            client = _dask_default_client()
+        except ValueError:
+            # No dask client found; continue
+            pass
+        else:
+            return client.loop
+
     if not _io_loops:
         loop = IOLoop()
         thread = threading.Thread(target=loop.start)
@@ -109,7 +122,79 @@ class RefCounter:
     __repr__ = __str__
 
 
-class Stream(object):
+class APIRegisterMixin(object):
+
+    @classmethod
+    def register_api(cls, modifier=identity, attribute_name=None):
+        """ Add callable to Stream API
+
+        This allows you to register a new method onto this class.  You can use
+        it as a decorator.::
+
+            >>> @Stream.register_api()
+            ... class foo(Stream):
+            ...     ...
+
+            >>> Stream().foo(...)  # this works now
+
+        It attaches the callable as a normal attribute to the class object.  In
+        doing so it respects inheritance (all subclasses of Stream will also
+        get the foo attribute).
+
+        By default callables are assumed to be instance methods.  If you like
+        you can include modifiers to apply before attaching to the class as in
+        the following case where we construct a ``staticmethod``.
+
+            >>> @Stream.register_api(staticmethod)
+            ... class foo(Stream):
+            ...     ...
+
+            >>> Stream.foo(...)  # Foo operates as a static method
+
+        You can also provide an optional ``attribute_name`` argument to control
+        the name of the attribute your callable will be attached as.
+
+            >>> @Stream.register_api(attribute_name="bar")
+            ... class foo(Stream):
+            ...     ...
+
+            >> Stream().bar(...)  # foo was actually attached as bar
+        """
+        def _(func):
+            @functools.wraps(func)
+            def wrapped(*args, **kwargs):
+                return func(*args, **kwargs)
+            name = attribute_name if attribute_name else func.__name__
+            setattr(cls, name, modifier(wrapped))
+            return func
+        return _
+
+    @classmethod
+    def register_plugin_entry_point(cls, entry_point, modifier=identity):
+        if hasattr(cls, entry_point.name):
+            raise ValueError(
+                f"Can't add {entry_point.name} from {entry_point.module_name} "
+                f"to {cls.__name__}: duplicate method name."
+            )
+
+        def stub(*args, **kwargs):
+            """ Entrypoints-based streamz plugin. Will be loaded on first call. """
+            node = entry_point.load()
+            if not issubclass(node, Stream):
+                raise TypeError(
+                    f"Error loading {entry_point.name} "
+                    f"from module {entry_point.module_name}: "
+                    f"{node.__class__.__name__} must be a subclass of Stream"
+                )
+            if getattr(cls, entry_point.name).__name__ == "stub":
+                cls.register_api(
+                    modifier=modifier, attribute_name=entry_point.name
+                )(node)
+            return node(*args, **kwargs)
+        cls.register_api(modifier=modifier, attribute_name=entry_point.name)(stub)
+
+
+class Stream(APIRegisterMixin):
     """ A Stream is an infinite sequence of data.
 
     Streams subscribe to each other passing and transforming data between them.
@@ -164,11 +249,16 @@ class Stream(object):
 
     def __init__(self, upstream=None, upstreams=None, stream_name=None,
                  loop=None, asynchronous=None, ensure_io_loop=False):
+        self.name = stream_name
         self.downstreams = OrderedWeakrefSet()
+        self.current_value = None
+        self.current_metadata = None
         if upstreams is not None:
             self.upstreams = list(upstreams)
-        else:
+        elif upstream is not None:
             self.upstreams = [upstream]
+        else:
+            self.upstreams = []
 
         self._set_asynchronous(asynchronous)
         self._set_loop(loop)
@@ -181,8 +271,6 @@ class Stream(object):
             if upstream:
                 upstream.downstreams.add(self)
 
-        self.name = stream_name
-
     def _set_loop(self, loop):
         self.loop = None
         if loop is not None:
@@ -238,10 +326,7 @@ class Stream(object):
     def _add_upstream(self, upstream):
         """Add upstream to current upstreams, this method is overridden for
         classes which handle stream specific buffers/caches"""
-        if self.upstreams == [None]:
-            self.upstreams[0] = upstream
-        else:
-            self.upstreams.append(upstream)
+        self.upstreams.append(upstream)
 
     def _add_downstream(self, downstream):
         """Add downstream to current downstreams"""
@@ -254,51 +339,18 @@ class Stream(object):
     def _remove_upstream(self, upstream):
         """Remove upstream from current upstreams, this method is overridden for
         classes which handle stream specific buffers/caches"""
-        if len(self.upstreams) == 1:
-            self.upstreams[0] = [None]
-        else:
-            self.upstreams.remove(upstream)
-
-    @classmethod
-    def register_api(cls, modifier=identity):
-        """ Add callable to Stream API
-
-        This allows you to register a new method onto this class.  You can use
-        it as a decorator.::
-
-            >>> @Stream.register_api()
-            ... class foo(Stream):
-            ...     ...
-
-            >>> Stream().foo(...)  # this works now
-
-        It attaches the callable as a normal attribute to the class object.  In
-        doing so it respsects inheritance (all subclasses of Stream will also
-        get the foo attribute).
-
-        By default callables are assumed to be instance methods.  If you like
-        you can include modifiers to apply before attaching to the class as in
-        the following case where we construct a ``staticmethod``.
-
-            >>> @Stream.register_api(staticmethod)
-            ... class foo(Stream):
-            ...     ...
-
-            >>> Stream.foo(...)  # Foo operates as a static method
-        """
-        def _(func):
-            @functools.wraps(func)
-            def wrapped(*args, **kwargs):
-                return func(*args, **kwargs)
-            setattr(cls, func.__name__, modifier(wrapped))
-            return func
-        return _
+        self.upstreams.remove(upstream)
 
     def start(self):
         """ Start any upstream sources """
         for upstream in self.upstreams:
             upstream.start()
 
+    def stop(self):
+        """ Stop upstream sources """
+        for upstream in self.upstreams:
+            upstream.stop()
+
     def __str__(self):
         s_list = []
         if self.name:
@@ -314,8 +366,6 @@ class Stream(object):
                     s = str(at)
                 elif hasattr(at, '__name__'):
                     s = getattr(self, m).__name__
-                elif hasattr(at.__class__, '__name__'):
-                    s = getattr(self, m).__class__.__name__
                 else:
                     s = None
             if s:
@@ -333,10 +383,10 @@ class Stream(object):
 
     __repr__ = __str__
 
-    def _ipython_display_(self, **kwargs):
+    def _ipython_display_(self, **kwargs):  # pragma: no cover
         try:
             from ipywidgets import Output
-            import IPython
+            from IPython.core.interactiveshell import InteractiveShell
         except ImportError:
             if hasattr(self, '_repr_html_'):
                 return self._repr_html_()
@@ -350,8 +400,10 @@ class Stream(object):
             if output is None:
                 return
             with output:
-                IPython.display.clear_output(wait=True)
-                IPython.display.display(val)
+                content, *_ = InteractiveShell.instance().display_formatter.format(val)
+                output.outputs = ({'output_type': 'display_data',
+                      'data': content,
+                      'metadata': {}},)
 
         s = self.map(update_cell)
         _html_update_streams.add(s)
@@ -388,6 +440,8 @@ class Stream(object):
             A reference counter used to check when data is done
 
         """
+        self.current_value = x
+        self.current_metadata = metadata
         if metadata:
             self._retain_refs(metadata, len(self.downstreams))
         else:
@@ -447,7 +501,7 @@ class Stream(object):
             sync(self.loop, _)
 
     def update(self, x, who=None, metadata=None):
-        self._emit(x, metadata=metadata)
+        return self._emit(x, metadata=metadata)
 
     def gather(self):
         """ This is a no-op for core streamz
@@ -481,8 +535,10 @@ class Stream(object):
 
     @property
     def upstream(self):
-        if len(self.upstreams) != 1:
+        if len(self.upstreams) > 1:
             raise ValueError("Stream has multiple upstreams")
+        elif len(self.upstreams) == 0:
+            return None
         else:
             return self.upstreams[0]
 
@@ -493,8 +549,8 @@ class Stream(object):
         if streams is None:
             streams = self.upstreams
         for upstream in list(streams):
-            upstream.downstreams.remove(self)
-            self.upstreams.remove(upstream)
+            upstream._remove_downstream(self)
+            self._remove_upstream(upstream)
 
     def scatter(self, **kwargs):
         from .dask import scatter
@@ -620,48 +676,6 @@ class Stream(object):
                 m['ref'].release(n)
 
 
-@Stream.register_api()
-class sink(Stream):
-    """ Apply a function on every element
-
-    Examples
-    --------
-    >>> source = Stream()
-    >>> L = list()
-    >>> source.sink(L.append)
-    >>> source.sink(print)
-    >>> source.sink(print)
-    >>> source.emit(123)
-    123
-    123
-    >>> L
-    [123]
-
-    See Also
-    --------
-    map
-    Stream.sink_to_list
-    """
-    _graphviz_shape = 'trapezium'
-
-    def __init__(self, upstream, func, *args, **kwargs):
-        self.func = func
-        # take the stream specific kwargs out
-        stream_name = kwargs.pop("stream_name", None)
-        self.kwargs = kwargs
-        self.args = args
-
-        Stream.__init__(self, upstream, stream_name=stream_name)
-        _global_sinks.add(self)
-
-    def update(self, x, who=None, metadata=None):
-        result = self.func(x, *self.args, **self.kwargs)
-        if gen.isawaitable(result):
-            return result
-        else:
-            return []
-
-
 @Stream.register_api()
 class map(Stream):
     """ Apply a function to every element in the stream
@@ -949,6 +963,19 @@ class slice(Stream):
 class partition(Stream):
     """ Partition stream into tuples of equal size
 
+    Parameters
+    ----------
+    n: int
+        Maximum partition size
+    timeout: int or float, optional
+        Number of seconds after which a partition will be emitted,
+        even if its size is less than ``n``. If ``None`` (default),
+        a partition will be emitted only when its size reaches ``n``.
+    key: hashable or callable, optional
+        Emit items with the same key together as a separate partition.
+        If ``key`` is callable, partition will be identified by ``key(x)``,
+        otherwise by ``x[key]``. Defaults to ``None``.
+
     Examples
     --------
     >>> source = Stream()
@@ -958,26 +985,165 @@ class partition(Stream):
     (0, 1, 2)
     (3, 4, 5)
     (6, 7, 8)
+
+    >>> source = Stream()
+    >>> source.partition(2, key=lambda x: x % 2).sink(print)
+    >>> for i in range(4):
+    ...     source.emit(i)
+    (0, 2)
+    (1, 3)
+
+    >>> from time import sleep
+    >>> source = Stream()
+    >>> source.partition(5, timeout=1).sink(print)
+    >>> for i in range(3):
+    ...     source.emit(i)
+    >>> sleep(1)
+    (0, 1, 2)
     """
     _graphviz_shape = 'diamond'
 
-    def __init__(self, upstream, n, **kwargs):
+    def __init__(self, upstream, n, timeout=None, key=None, **kwargs):
         self.n = n
-        self._buffer = []
-        self.metadata_buffer = []
+        self._timeout = timeout
+        self._key = key
+        self._buffer = defaultdict(lambda: [])
+        self._metadata_buffer = defaultdict(lambda: [])
+        self._callbacks = {}
+        kwargs["ensure_io_loop"] = True
         Stream.__init__(self, upstream, **kwargs)
 
+    def _get_key(self, x):
+        if self._key is None:
+            return None
+        if callable(self._key):
+            return self._key(x)
+        return x[self._key]
+
+    @gen.coroutine
+    def _flush(self, key):
+        result, self._buffer[key] = self._buffer[key], []
+        metadata_result, self._metadata_buffer[key] = self._metadata_buffer[key], []
+        yield self._emit(tuple(result), list(metadata_result))
+        self._release_refs(metadata_result)
+
+    @gen.coroutine
+    def update(self, x, who=None, metadata=None):
+        self._retain_refs(metadata)
+        key = self._get_key(x)
+        buffer = self._buffer[key]
+        metadata_buffer = self._metadata_buffer[key]
+        buffer.append(x)
+        if isinstance(metadata, list):
+            metadata_buffer.extend(metadata)
+        else:
+            metadata_buffer.append(metadata)
+        if len(buffer) == self.n:
+            if self._timeout is not None and self.n > 1:
+                self._callbacks[key].cancel()
+            yield self._flush(key)
+            return
+        if len(buffer) == 1 and self._timeout is not None:
+            self._callbacks[key] = self.loop.call_later(
+                self._timeout, self._flush, key
+            )
+
+
+@Stream.register_api()
+class partition_unique(Stream):
+    """
+    Partition stream elements into groups of equal size with unique keys only.
+
+    Parameters
+    ----------
+    n: int
+        Number of (unique) elements to pass through as a group.
+    key: Union[Hashable, Callable[[Any], Hashable]]
+        Callable that accepts a stream element and returns a unique, hashable
+        representation of the incoming data (``key(x)``), or a hashable that gets
+        the corresponding value of a stream element (``x[key]``). For example,
+        ``key=lambda x: x["a"]`` would allow only elements with unique ``"a"`` values
+        to pass through.
+
+        .. note:: By default, we simply use the element object itself as the key,
+            so that object must be hashable. If that's not the case, a non-default
+            key must be provided.
+
+    keep: str
+        Which element to keep in the case that a unique key is already found
+        in the group. If "first", keep element from the first occurrence of a given
+        key; if "last", keep element from the most recent occurrence. Note that
+        relative ordering of *elements* is preserved in the data passed through,
+        and not ordering of *keys*.
+    **kwargs
+
+    Examples
+    --------
+    >>> source = Stream()
+    >>> stream = source.partition_unique(n=3, keep="first").sink(print)
+    >>> eles = [1, 2, 1, 3, 1, 3, 3, 2]
+    >>> for ele in eles:
+    ...     source.emit(ele)
+    (1, 2, 3)
+    (1, 3, 2)
+
+    >>> source = Stream()
+    >>> stream = source.partition_unique(n=3, keep="last").sink(print)
+    >>> eles = [1, 2, 1, 3, 1, 3, 3, 2]
+    >>> for ele in eles:
+    ...     source.emit(ele)
+    (2, 1, 3)
+    (1, 3, 2)
+
+    >>> source = Stream()
+    >>> stream = source.partition_unique(n=3, key=lambda x: len(x), keep="last").sink(print)
+    >>> eles = ["f", "fo", "f", "foo", "f", "foo", "foo", "fo"]
+    >>> for ele in eles:
+    ...     source.emit(ele)
+    ('fo', 'f', 'foo')
+    ('f', 'foo', 'fo')
+    """
+    _graphviz_shape = "diamond"
+
+    def __init__(
+        self,
+        upstream,
+        n: int,
+        key: Union[Hashable, Callable[[Any], Hashable]] = identity,
+        keep: str = "first",  # Literal["first", "last"]
+        **kwargs
+    ):
+        self.n = n
+        self.key = key
+        self.keep = keep
+        self._buffer = {}
+        self._metadata_buffer = {}
+        Stream.__init__(self, upstream, **kwargs)
+
+    def _get_key(self, x):
+        if callable(self.key):
+            return self.key(x)
+        else:
+            return x[self.key]
+
     def update(self, x, who=None, metadata=None):
         self._retain_refs(metadata)
-        self._buffer.append(x)
-        if isinstance(metadata, list):
-            self.metadata_buffer.extend(metadata)
-        else:
-            self.metadata_buffer.append(metadata)
+        y = self._get_key(x)
+        if self.keep == "last":
+            # remove key if already present so that emitted value
+            # will reflect elements' actual relative ordering
+            self._buffer.pop(y, None)
+            self._metadata_buffer.pop(y, None)
+            self._buffer[y] = x
+            self._metadata_buffer[y] = metadata
+        else:  # self.keep == "first"
+            if y not in self._buffer:
+                self._buffer[y] = x
+                self._metadata_buffer[y] = metadata
         if len(self._buffer) == self.n:
-            result, self._buffer = self._buffer, []
-            metadata_result, self.metadata_buffer = self.metadata_buffer, []
-            ret = self._emit(tuple(result), list(metadata_result))
+            result, self._buffer = tuple(self._buffer.values()), {}
+            metadata_result, self._metadata_buffer = list(self._metadata_buffer.values()), {}
+            ret = self._emit(result, metadata_result)
             self._release_refs(metadata_result)
             return ret
         else:
@@ -1057,7 +1223,8 @@ class timed_window(Stream):
         self.metadata_buffer = []
         self.last = gen.moment
 
-        Stream.__init__(self, upstream, ensure_io_loop=True, **kwargs)
+        kwargs["ensure_io_loop"] = True
+        Stream.__init__(self, upstream, **kwargs)
 
         self.loop.add_callback(self.cb)
 
@@ -1079,6 +1246,125 @@ class timed_window(Stream):
             yield gen.sleep(self.interval)
 
 
+@Stream.register_api()
+class timed_window_unique(Stream):
+    """
+    Emit a group of elements with unique keys every ``interval`` seconds.
+
+    Parameters
+    ----------
+    interval: Union[int, str]
+        Number of seconds over which to group elements, or a ``pandas``-style
+        duration string that can be converted into seconds.
+    key: Union[Hashable, Callable[[Any], Hashable]]
+        Callable that accepts a stream element and returns a unique, hashable
+        representation of the incoming data (``key(x)``), or a hashable that gets
+        the corresponding value of a stream element (``x[key]``). For example, both
+        ``key=lambda x: x["a"]`` and ``key="a"`` would allow only elements with unique
+        ``"a"`` values to pass through.
+
+        .. note:: By default, we simply use the element object itself as the key,
+            so that object must be hashable. If that's not the case, a non-default
+            key must be provided.
+
+    keep: str
+        Which element to keep in the case that a unique key is already found
+        in the group. If "first", keep element from the first occurrence of a given
+        key; if "last", keep element from the most recent occurrence. Note that
+        relative ordering of *elements* is preserved in the data passed through,
+        and not ordering of *keys*.
+
+    Examples
+    --------
+    >>> source = Stream()
+
+    Get unique hashable elements in a window, keeping just the first occurrence:
+    >>> stream = source.timed_window_unique(interval=1.0, keep="first").sink(print)
+    >>> for ele in [1, 2, 3, 3, 2, 1]:
+    ...     source.emit(ele)
+    ()
+    (1, 2, 3)
+    ()
+
+    Get unique hashable elements in a window, keeping just the last occurrence:
+    >>> stream = source.timed_window_unique(interval=1.0, keep="last").sink(print)
+    >>> for ele in [1, 2, 3, 3, 2, 1]:
+    ...     source.emit(ele)
+    ()
+    (3, 2, 1)
+    ()
+
+    Get unique elements in a window by (string) length, keeping just the first occurrence:
+    >>> stream = source.timed_window_unique(interval=1.0, key=len, keep="first")
+    >>> for ele in ["f", "b", "fo", "ba", "foo", "bar"]:
+    ...     source.emit(ele)
+    ()
+    ('f', 'fo', 'foo')
+    ()
+
+    Get unique elements in a window by (string) length, keeping just the last occurrence:
+    >>> stream = source.timed_window_unique(interval=1.0, key=len, keep="last")
+    >>> for ele in ["f", "b", "fo", "ba", "foo", "bar"]:
+    ...     source.emit(ele)
+    ()
+    ('b', 'ba', 'bar')
+    ()
+    """
+    _graphviz_shape = "octagon"
+
+    def __init__(
+        self,
+        upstream,
+        interval: Union[int, str],
+        key: Union[Hashable, Callable[[Any], Hashable]] = identity,
+        keep: str = "first",  # Literal["first", "last"]
+        **kwargs
+    ):
+        self.interval = convert_interval(interval)
+        self.key = key
+        self.keep = keep
+        self._buffer = {}
+        self._metadata_buffer = {}
+        self.last = gen.moment
+        kwargs["ensure_io_loop"] = True
+        Stream.__init__(self, upstream, **kwargs)
+        self.loop.add_callback(self.cb)
+
+    def _get_key(self, x):
+        if callable(self.key):
+            return self.key(x)
+        else:
+            return x[self.key]
+
+    def update(self, x, who=None, metadata=None):
+        self._retain_refs(metadata)
+        y = self._get_key(x)
+        if self.keep == "last":
+            # remove key if already present so that emitted value
+            # will reflect elements' actual relative ordering
+            self._buffer.pop(y, None)
+            self._metadata_buffer.pop(y, None)
+            self._buffer[y] = x
+            self._metadata_buffer[y] = metadata
+        else:  # self.keep == "first"
+            if y not in self._buffer:
+                self._buffer[y] = x
+                self._metadata_buffer[y] = metadata
+        return self.last
+
+    @gen.coroutine
+    def cb(self):
+        while True:
+            result, self._buffer = tuple(self._buffer.values()), {}
+            metadata_result, self._metadata_buffer = list(self._metadata_buffer.values()), {}
+            # TODO: figure out why metadata_result is handled differently here...
+            m = [m for ml in metadata_result for m in ml]
+            self.last = self._emit(result, m)
+            self._release_refs(m)
+            yield self.last
+            yield gen.sleep(self.interval)
+
+
 @Stream.register_api()
 class delay(Stream):
     """ Add a time delay to results """
@@ -1088,7 +1374,8 @@ class delay(Stream):
         self.interval = convert_interval(interval)
         self.queue = Queue()
 
-        Stream.__init__(self, upstream, ensure_io_loop=True, **kwargs)
+        kwargs["ensure_io_loop"] = True
+        Stream.__init__(self, upstream,**kwargs)
 
         self.loop.add_callback(self.cb)
 
@@ -1126,7 +1413,8 @@ class rate_limit(Stream):
         self.interval = convert_interval(interval)
         self.next = 0
 
-        Stream.__init__(self, upstream, ensure_io_loop=True, **kwargs)
+        kwargs["ensure_io_loop"] = True
+        Stream.__init__(self, upstream, **kwargs)
 
     @gen.coroutine
     def update(self, x, who=None, metadata=None):
@@ -1151,7 +1439,8 @@ class buffer(Stream):
     def __init__(self, upstream, n, **kwargs):
         self.queue = Queue(maxsize=n)
 
-        Stream.__init__(self, upstream, ensure_io_loop=True, **kwargs)
+        kwargs["ensure_io_loop"] = True
+        Stream.__init__(self, upstream, **kwargs)
 
         self.loop.add_callback(self.cb)
 
@@ -1595,7 +1884,8 @@ class latest(Stream):
         self.next = []
         self.next_metadata = None
 
-        Stream.__init__(self, upstream, ensure_io_loop=True, **kwargs)
+        kwargs["ensure_io_loop"] = True
+        Stream.__init__(self, upstream, **kwargs)
 
         self.loop.add_callback(self.cb)
 
@@ -1651,7 +1941,8 @@ class to_kafka(Stream):
         self.topic = topic
         self.producer = ck.Producer(producer_config)
 
-        Stream.__init__(self, upstream, ensure_io_loop=True, **kwargs)
+        kwargs["ensure_io_loop"] = True
+        Stream.__init__(self, upstream, **kwargs)
         self.stopped = False
         self.polltime = 0.2
         self.loop.add_callback(self.poll)
