@@ -1,12 +1,11 @@
-from __future__ import division, print_function
+import asyncio
 
 import operator
 from collections import OrderedDict
-from time import time
 import numpy as np
 import pandas as pd
 import toolz
-from tornado.ioloop import IOLoop
+
 from tornado import gen
 
 from ..collection import Streaming, _stream_types, OperatorMixin
@@ -246,7 +245,7 @@ class _DataFrameMixin(object):
         o = set(dir(type(self)))
         o.update(self.__dict__)
         o.update(c for c in self.columns
-                 if (isinstance(c, pd.compat.string_types) and pd.compat.isidentifier(c)))
+                 if (isinstance(c, str) and c.isidentifier()))
         return list(o)
 
     def assign(self, **kwargs):
@@ -327,6 +326,10 @@ class DataFrame(Frame, _DataFrameMixin):
                 example = kwargs.get('example')
             elif len(args) > 1:
                 example = args[1]
+            if callable(example):
+                example = example()
+                kwargs["example"] = example
+
             self._subtype = get_base_frame_type(self.__class__.__name__,
                                                 is_dataframe_like, example)
             super(DataFrame, self).__init__(*args, **kwargs)
@@ -338,6 +341,15 @@ class DataFrame(Frame, _DataFrameMixin):
             raise IndexError("Input expected to have columns %s, got %s" %
                              (self.example.columns, x.columns))
 
+    @property
+    def plot(self):
+        try:
+            # import has side-effect of attaching .hvplot attribute
+            import hvplot.streamz  # # noqa: F401
+        except ImportError as err:  # pragma: no cover
+            raise ImportError("Streamz dataframe plotting requires hvplot") from err
+        return self.hvplot
+
 
 class _SeriesMixin(object):
     @property
@@ -802,10 +814,37 @@ class WindowedGroupBy(GroupBy):
         return Streaming(outstream, example, stream_type=stream_type)
 
 
-def _random_df(tup):
-    last, now, freq = tup
-    index = pd.date_range(start=(last + freq.total_seconds()) * 1e9,
-                          end=now * 1e9, freq=freq)
+def random_datapoint(now=None, **kwargs):
+    """Example of querying a single current value"""
+    if now is None:
+        now = pd.Timestamp.now()
+    return pd.DataFrame(
+        {'a': np.random.random(1)}, index=[now])
+
+
+def random_datablock(last, now, **kwargs):
+    """
+    Example of querying over a time range since last update
+
+    Parameters
+    ----------
+    last: pd.Timestamp
+        Time of previous call to this function.
+    now: pd.Timestamp
+        Current time.
+    freq: pd.Timedelta, optional
+        The time interval between individual records to be returned.
+        For good throughput, should be much smaller than the
+        interval at which this function is called.
+
+    Returns a pd.DataFrame with random values where:
+
+    The x column is uniformly distributed.
+    The y column is Poisson distributed.
+    The z column is normally distributed.
+    """
+    freq = kwargs.get("freq", pd.Timedelta("100ms"))
+    index = pd.date_range(start=last + freq, end=now, freq=freq)
 
     df = pd.DataFrame({'x': np.random.random(len(index)),
                        'y': np.random.poisson(size=len(index)),
@@ -814,48 +853,57 @@ def _random_df(tup):
     return df
 
 
-class Random(DataFrame):
-    """ A streaming dataframe of random data
-
-    The x column is uniformly distributed.
-    The y column is poisson distributed.
-    The z column is normally distributed.
-
-    This class is experimental and will likely be removed in the future
+@DataFrame.register_api(staticmethod, "from_periodic")
+class PeriodicDataFrame(DataFrame):
+    """A streaming dataframe using the asyncio ioloop to poll a callback fn
 
     Parameters
     ----------
-    freq: timedelta
-        The time interval between records
+    datafn: callable
+        Callback function accepting **kwargs and returning a
+        pd.DataFrame.  kwargs will include at least
+        'last' (pd.Timestamp.now() when datafn was last invoked), and
+        'now' (current pd.Timestamp.now()).
     interval: timedelta
-        The time interval between new dataframes, should be significantly
-        larger than freq
+        The time interval between new dataframes.
+    dask: boolean
+        If true, uses a DaskStream instead of a regular Source.
+    **kwargs:
+        Optional keyword arguments to be passed into the callback function.
+
+    By default, returns a three-column random pd.DataFrame generated
+    by the 'random_datablock' function.
 
     Example
     -------
-    >>> source = Random(freq='100ms', interval='1s')  # doctest: +SKIP
+    >>> df = PeriodicDataFrame(interval='1s', datafn=random_datapoint)  # doctest: +SKIP
     """
 
-    def __init__(self, freq='100ms', interval='500ms', dask=False):
+    def __init__(self, datafn=random_datablock, interval='500ms', dask=False,
+                 start=True, **kwargs):
         if dask:
             from streamz.dask import DaskStream
             source = DaskStream()
-            loop = source.loop
         else:
             source = Source()
-            loop = IOLoop.current()
-        self.freq = pd.Timedelta(freq)
+        self.loop = source.loop
         self.interval = pd.Timedelta(interval).total_seconds()
         self.source = source
-        self.continue_ = [True]
+        self.continue_ = [False]  # like the oppose of self.stopped
+        self.kwargs = kwargs
 
-        stream = self.source.map(_random_df)
-        example = _random_df((time(), time(), self.freq))
+        stream = self.source.map(lambda x: datafn(**x, **kwargs))
+        example = datafn(last=pd.Timestamp.now(), now=pd.Timestamp.now(), **kwargs)
 
-        super(Random, self).__init__(stream, example)
+        super(PeriodicDataFrame, self).__init__(stream, example)
+        if start:
+            self.start()
 
-        loop.add_callback(self._cb, self.interval, self.freq, self.source,
-                          self.continue_)
+    def start(self):
+        if not self.continue_[0]:
+            self.continue_[0] = True
+            self.loop.add_callback(self._cb, self.interval, self.source,
+                                   self.continue_)
 
     def __del__(self):
         self.stop()
@@ -864,16 +912,36 @@ class Random(DataFrame):
         self.continue_[0] = False
 
     @staticmethod
-    @gen.coroutine
-    def _cb(interval, freq, source, continue_):
-        last = time()
+    async def _cb(interval, source, continue_):
+        last = pd.Timestamp.now()
         while continue_[0]:
-            yield gen.sleep(interval)
-            now = time()
-            yield source._emit((last, now, freq))
+            await gen.sleep(interval)
+            now = pd.Timestamp.now()
+            await asyncio.gather(*source._emit(dict(last=last, now=now)))
             last = now
 
 
+@DataFrame.register_api(staticmethod, "random")
+class Random(PeriodicDataFrame):
+    """PeriodicDataFrame providing random values by default
+
+    Accepts same parameters as PeriodicDataFrame, plus
+    `freq`, a string that will be converted to a pd.Timedelta
+    and passed to the 'datafn'.
+
+    Useful mainly for examples and docs.
+
+    Example
+    -------
+    >>> source = Random(freq='100ms', interval='1s')  # doctest: +SKIP
+    """
+
+    def __init__(self, freq='100ms', interval='500ms', dask=False,
+                 start=True, datafn=random_datablock):
+        super(Random, self).__init__(datafn, interval, dask, start,
+                                     freq=pd.Timedelta(freq))
+
+
 _stream_types['streaming'].append((is_dataframe_like, DataFrame))
 _stream_types['streaming'].append((is_index_like, Index))
 _stream_types['streaming'].append((is_series_like, Series))
