@@ -11,7 +11,7 @@ import pandas as pd
 from tornado import gen
 
 from streamz import Stream
-from streamz.utils_test import gen_test
+from streamz.utils_test import gen_test, wait_for
 from streamz.dataframe import DataFrame, Series, DataFrames, Aggregation
 import streamz.dataframe as sd
 from streamz.dask import DaskStream
@@ -231,6 +231,8 @@ def test_index(stream):
     a.emit(df)
     a.emit(df)
 
+    wait_for(lambda: len(L) > 1, timeout=2, period=0.05)
+
     assert_eq(L[0], df.index + 5)
     assert_eq(L[1], df.index + 5)
 
@@ -555,7 +557,6 @@ def test_gc():
     a = DataFrame({'volatility': sdf.x.rolling('100ms').var(),
                             'sub': sdf.x - sdf.x.rolling('100ms').mean()})
     n = len(sdf.stream.downstreams)
-    yield gen.sleep(0.1)
     a = DataFrame({'volatility': sdf.x.rolling('100ms').var(),
                             'sub': sdf.x - sdf.x.rolling('100ms').mean()})
     yield gen.sleep(0.1)
@@ -564,6 +565,7 @@ def test_gc():
     yield gen.sleep(0.1)
     a = DataFrame({'volatility': sdf.x.rolling('100ms').var(),
                             'sub': sdf.x - sdf.x.rolling('100ms').mean()})
+    yield gen.sleep(0.1)
 
     assert len(sdf.stream.downstreams) == n
     del a
@@ -755,12 +757,11 @@ def test_groupby_windowing_value(func, value, getter, grouper, indexer):
     assert len(L) == 5
 
     first = df.iloc[:diff]
-    lost = first[first.index.min() + value:]
-    first = first.iloc[len(lost):]
+    first = first[first.index.max() - value + pd.Timedelta('1ns'):]
 
     assert_eq(L[0], f(first))
 
-    last = df.loc[index.max() - value + pd.Timedelta('1s'):]
+    last = df.loc[index.max() - value + pd.Timedelta('1ns'):]
 
     assert_eq(L[-1], f(last))
 
@@ -815,6 +816,42 @@ def test_groupby_windowing_n(func, n, getter, grouper, indexer):
     assert_eq(L[-1], f(last))
 
 
+def test_windowing_value_empty_intermediate_index(stream):
+    def preprocess(df):
+        mask = df["amount"] == 5
+        df = df.loc[mask]
+        return df
+
+    source = stream.map(preprocess)
+
+    example = pd.DataFrame({"amount":[]})
+    sdf = DataFrame(stream=source, example=example)
+
+    output = sdf.window("2h").amount.sum().stream.gather().sink_to_list()
+
+    stream.emit(pd.DataFrame({"amount": [1, 2, 3]}, index=[pd.Timestamp("2050-01-01 00:00:00"),
+                                                           pd.Timestamp("2050-01-01 01:00:00"),
+                                                           pd.Timestamp("2050-01-01 02:00:00")]))
+
+    stream.emit(pd.DataFrame({"amount": [5, 5, 5]}, index=[pd.Timestamp("2050-01-01 03:00:00"),
+                                                           pd.Timestamp("2050-01-01 04:00:00"),
+                                                           pd.Timestamp("2050-01-01 05:00:00")]))
+
+    stream.emit(pd.DataFrame({"amount": [4, 5, 6]}, index=[pd.Timestamp("2050-01-01 06:00:00"),
+                                                           pd.Timestamp("2050-01-01 07:00:00"),
+                                                           pd.Timestamp("2050-01-01 08:00:00")]))
+
+    stream.emit(pd.DataFrame({"amount": [1, 2, 3]}, index=[pd.Timestamp("2050-01-01 09:00:00"),
+                                                           pd.Timestamp("2050-01-01 10:00:00"),
+                                                           pd.Timestamp("2050-01-01 11:00:00")]))
+
+    stream.emit(pd.DataFrame({"amount": [5, 5, 5]}, index=[pd.Timestamp("2050-01-01 12:00:00"),
+                                                           pd.Timestamp("2050-01-01 13:00:00"),
+                                                           pd.Timestamp("2050-01-01 14:00:00")]))
+
+    assert_eq(output, [0, 10, 5, 5, 10])
+
+
 def test_window_full():
     df = pd.DataFrame({'x': np.arange(10, dtype=float), 'y': [1.0, 2.0] * 5})
 
@@ -972,3 +1009,10 @@ def test_windowed_groupby_aggs_with_start_state(stream):
     stream.emit(df)
     out_df1 = pd.DataFrame({'name':['Alice', 'Bob', 'Linda', 'Tom'], 'amount':[50.0, 550.0, 100.0, 150.0]})
     assert_eq(output1[-1][1].reset_index(), out_df1)
+
+
+def test_dir(stream):
+    example = pd.DataFrame({'name': [], 'amount': []})
+    sdf = DataFrame(stream, example=example)
+    assert 'name' in dir(sdf)
+    assert 'amount' in dir(sdf)
