@@ -5,21 +5,20 @@ import json
 import operator
 from operator import add
 import os
-from time import time, sleep
+from time import sleep
 import sys
 
 import pytest
 
-from tornado import gen
 from tornado.queues import Queue
 from tornado.ioloop import IOLoop
 
 import streamz as sz
 
-from streamz import Stream, RefCounter
-from streamz.sources import sink_to_file, PeriodicCallback
+from streamz import RefCounter
+from streamz.sources import sink_to_file
 from streamz.utils_test import (inc, double, gen_test, tmpfile, captured_logger,   # noqa: F401
-        clean, await_for, metadata)  # noqa: F401
+        clean, await_for, metadata, wait_for)  # noqa: F401
 from distributed.utils_test import loop   # noqa: F401
 
 
@@ -164,6 +163,102 @@ def test_partition():
     assert L == [(0, 1), (2, 3), (4, 5), (6, 7), (8, 9)]
 
 
+@pytest.mark.parametrize(
+    "n,key,keep,elements,exp_result",
+    [
+        (3, sz.identity, "first", [1, 2, 1, 3, 1, 3, 3, 2], [(1, 2, 3), (1, 3, 2)]),
+        (3, sz.identity, "last", [1, 2, 1, 3, 1, 3, 3, 2], [(2, 1, 3), (1, 3, 2)]),
+        (
+            3,
+            len,
+            "last",
+            ["f", "fo", "f", "foo", "f", "foo", "foo", "fo"],
+            [("fo", "f", "foo"), ("f", "foo", "fo")],
+        ),
+        (
+            2,
+            "id",
+            "first",
+            [{"id": 0, "foo": "bar"}, {"id": 0, "foo": "baz"}, {"id": 1, "foo": "bat"}],
+            [({"id": 0, "foo": "bar"}, {"id": 1, "foo": "bat"})],
+        ),
+        (
+            2,
+            "id",
+            "last",
+            [{"id": 0, "foo": "bar"}, {"id": 0, "foo": "baz"}, {"id": 1, "foo": "bat"}],
+            [({"id": 0, "foo": "baz"}, {"id": 1, "foo": "bat"})],
+        ),
+    ]
+)
+def test_partition_unique(n, key, keep, elements, exp_result):
+    source = Stream()
+    L = source.partition_unique(n, key, keep).sink_to_list()
+    for ele in elements:
+        source.emit(ele)
+
+    assert L == exp_result
+
+
+def test_partition_timeout():
+    source = Stream()
+    L = source.partition(10, timeout=0.01).sink_to_list()
+
+    for i in range(5):
+        source.emit(i)
+
+    sleep(0.1)
+
+    assert L == [(0, 1, 2, 3, 4)]
+
+
+def test_partition_timeout_cancel():
+    source = Stream()
+    L = source.partition(3, timeout=0.1).sink_to_list()
+
+    for i in range(3):
+        source.emit(i)
+
+    sleep(0.09)
+    source.emit(3)
+    sleep(0.02)
+
+    assert L == [(0, 1, 2)]
+
+    sleep(0.09)
+
+    assert L == [(0, 1, 2), (3,)]
+
+
+def test_partition_key():
+    source = Stream()
+    L = source.partition(2, key=0).sink_to_list()
+
+    for i in range(4):
+        source.emit((i % 2, i))
+
+    assert L == [((0, 0), (0, 2)), ((1, 1), (1, 3))]
+
+
+def test_partition_key_callable():
+    source = Stream()
+    L = source.partition(2, key=lambda x: x % 2).sink_to_list()
+
+    for i in range(10):
+        source.emit(i)
+
+    assert L == [(0, 2), (1, 3), (4, 6), (5, 7)]
+
+
+def test_partition_size_one():
+    source = Stream()
+
+    source.partition(1, timeout=.01).sink(lambda x: None)
+
+    for i in range(10):
+        source.emit(i)
+
+
 def test_sliding_window():
     source = Stream()
     L = source.sliding_window(2).sink_to_list()
@@ -237,6 +332,53 @@ def test_backpressure():
     assert end - start >= 0.2
 
 
+@gen_test()
+def test_timed_window_unique():
+    tests = [
+        (0.05, sz.identity, "first", [1, 2, 1, 3, 1, 3, 3, 2], [(1, 2, 3)]),
+        (0.05, sz.identity, "last", [1, 2, 1, 3, 1, 3, 3, 2], [(1, 3, 2)]),
+        (
+            0.05,
+            len,
+            "last",
+            ["f", "fo", "f", "foo", "f", "foo", "foo", "fo"],
+            [("f", "foo", "fo")],
+        ),
+        (
+            0.05,
+            "id",
+            "first",
+            [{"id": 0, "foo": "bar"}, {"id": 1, "foo": "bat"}, {"id": 0, "foo": "baz"}],
+            [({"id": 0, "foo": "bar"}, {"id": 1, "foo": "bat"})],
+        ),
+        (
+            0.05,
+            "id",
+            "last",
+            [{"id": 0, "foo": "bar"}, {"id": 1, "foo": "bat"}, {"id": 0, "foo": "baz"}],
+            [({"id": 1, "foo": "bat"}, {"id": 0, "foo": "baz"})],
+        ),
+    ]
+    for interval, key, keep, elements, exp_result in tests:
+        source = Stream(asynchronous=True)
+        a = source.timed_window_unique(interval, key, keep)
+
+        assert a.loop is IOLoop.current()
+        L = a.sink_to_list()
+
+        for ele in elements:
+            yield source.emit(ele)
+        yield gen.sleep(a.interval)
+
+        assert L
+        assert all(wi in elements for window in L for wi in window)
+        assert sum(1 for window in L for _ in window) <= len(elements)
+        assert L == exp_result
+
+        yield gen.sleep(a.interval)
+        assert not L[-1]
+
+
 @gen_test()
 def test_timed_window():
     source = Stream(asynchronous=True)
@@ -261,10 +403,11 @@ def test_timed_window():
 
 @gen_test()
 def test_timed_window_ref_counts():
-    source = Stream()
+    source = Stream(asynchronous=True)
     _ = source.timed_window(0.01)
 
     ref1 = RefCounter()
+    assert str(ref1) == "<RefCounter count=0>"
     source.emit(1, metadata=[{'ref': ref1}])
     assert ref1.count == 1
     yield gen.sleep(0.05)
@@ -275,10 +418,17 @@ def test_timed_window_ref_counts():
     assert ref2.count == 1
 
 
+def test_mixed_async():
+    s1 = Stream(asynchronous=False)
+    with pytest.raises(ValueError):
+        Stream(asynchronous=True, upstream=s1)
+
+
+
 @gen_test()
 def test_timed_window_metadata():
     source = Stream()
-    L = metadata(source.timed_window(0.01)).sink_to_list()
+    L = metadata(source.timed_window(0.06)).sink_to_list()
 
     source.emit(0)
     source.emit(1, metadata=[{'v': 1}])
@@ -336,27 +486,11 @@ def test_sink_to_file():
         assert data == 'a\nb\n'
 
 
-def test_sink_with_args_and_kwargs():
-    L = dict()
-
-    def mycustomsink(elem, key, prefix=""):
-        key = prefix + key
-        if key not in L:
-            L[key] = list()
-        L[key].append(elem)
-
-    s = Stream()
-    s.sink(mycustomsink, "cat", "super")
-
-    s.emit(1)
-    s.emit(2)
-    assert L['supercat'] == [1, 2]
-
-
 @gen_test()
 def test_counter():
     counter = itertools.count()
-    source = PeriodicCallback(lambda: next(counter), 0.001, asynchronous=True)
+    source = Stream.from_periodic(lambda: next(counter), 0.001, asynchronous=True,
+                                  start=True)
     L = source.sink_to_list()
     yield gen.sleep(0.05)
 
@@ -1038,7 +1172,7 @@ def test_connect():
     # connect assumes this default behaviour
     # of stream initialization
     assert not source_downstream.downstreams
-    assert source_downstream.upstreams == [None]
+    assert source_downstream.upstreams == []
 
     # initialize the second stream to connect to
     source_upstream = Stream()
@@ -1575,3 +1709,31 @@ def test_buffer_after_timed_window():
 
 def test_buffer_after_sliding_window():
     Stream().sliding_window(1).buffer(1)
+
+
+def test_backpressure_connect_empty_stream():
+    @Stream.register_api()
+    class from_list(Stream):
+
+        def __init__(self, source, **kwargs):
+            self.source = source
+            super().__init__(ensure_io_loop=True, **kwargs)
+
+        def start(self):
+            self.stopped = False
+            self.loop.add_callback(self.run)
+
+        @gen.coroutine
+        def run(self):
+            while not self.stopped and len(self.source) > 0:
+                yield self._emit(self.source.pop(0))
+
+    source_list = [0, 1, 2, 3, 4]
+    source = Stream.from_list(source_list)
+    sout = Stream()
+    L = sout.rate_limit(1).sink_to_list()
+    source.connect(sout)
+    source.start()
+
+    wait_for(lambda: L == [0], 0.01)
+    assert len(source_list) > 0
